-- espio's battle mod script
local loaded = false
espio.BattleSetup = function()
	if not (CBW_Battle) then return end	-- battle isnt loaded
	if (loaded) then return end	-- already setup for battle? dont run
	local B = CBW_Battle
	local S = B.SkinVars
	
	-- espio's priority
	local function espio_battlePriority(p)
		if (p.actionstate == 2)
			-- espio should have no attacking power himself, the tornado does the damaging here
			local atk, def = 0, 2
			B.SetPriority(p, atk, def, nil, atk, def, "leaf swirl")
		else
			local spinjump = ((p.pflags & PF_JUMPED) and not (p.pflags & PF_NOJUMPDAMAGE))
			local spinning = (p.pflags & PF_SPINNING)
			
			if (spinjump or spinning)
				-- make espio have a drill spin instead of a spin attack
				local atk, def = 1, 1
				local text = (spinning) and "drill-spin attack" or "jumping drill-spin attack"
				B.SetPriority(p, atk, def, nil, atk, def, text)
			end
		end
	end
	
	-- separate function for leaf swirl functionality, just so I can return on the end conditions
	local function espio_leafSwirl(p)
		local mo = p.mo
		
		-- end conditions
		if (p.actiontime == espio.LeafSwirlDuration)	-- duration is up
		or (P_PlayerInPain(p))	-- we're in pain
			if not (P_PlayerInPain(p))	-- we wouldnt want to set any state here
				mo.state = (P_IsObjectOnGround(mo)) and S_PLAY_STND or S_PLAY_FALL
			end
			
			p.actionstate = 0
			mo.espio_leafswirl_parryhack = 3	-- need this here too
			return
		end
				
		if (mo.state != S_ESPIO_LEAFSWIRL) then mo.state = S_ESPIO_LEAFSWIRL end
		local frametics = {5, 3, 2}
				
		-- get frame
		local frame
		local tics = 0
		for i = 0, #frametics-1
			tics = $+frametics[i+1]
			if (p.actiontime <= tics)
				-- only reduce momentum before the final frame
				local moms = {"momx", "momy", "momz"}
				for i = 1, 3
					local mom = moms[i]
					mo[mom] = FixedMul($, espio.LeafSwirlMultiplier)
				end
						
				frame = i
				break
			end
		end
		if (frame == nil) then frame = 6 end	-- assume we're on the final frame
		espio.SetFrame(mo, frame)
				
		-- final frame is attack time
		if (frame == 6)
			mo.espio_leafswirl_parryhack = 3	-- we'll need this to get parried!
			
			-- spawn the tornado
			if (p.actiontime == tics+1)
				local tornado = P_SpawnMobjFromMobj(mo, 0, 0, FixedMul(mo.scale, espio.LeafSwirlTornadoSpawnOffset), MT_ESPIO_LEAFSWIRL_TORNADO)
				tornado.target = mo	-- we'll want to save this to know who spawned it
				
				-- make sure it's at least always on the floor
				local checkz = tornado.z
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = tornado.z + tornado.height
				end
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz > P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height))
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz < P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height))
					local newz = P_FloorzAtPos(mo.x, mo.y, mo.z, mo.height)
					if (mo.eflags & MFE_VERTICALFLIP)
						newz = P_CeilingzAtPos(mo.x, mo.y, mo.z, mo.height)
					end
					
					P_TeleportMove(tornado, tornado.x, tornado.y, newz)
				end
			end
			
			-- remove all momentum at this point
			mo.momx, mo.momy, mo.momz = 0, 0, 0
		end
	end
	local function espio_battleSpecial(mo, doaction, pressedaction)
		local p = mo.player
		local grounded = P_IsObjectOnGround(mo)
		
		-- ability names
		if (mo.espio_chargejump)
			p.actiontext = "Charge Thrust"
			p.actionrings = 10
		else
			p.actiontext = (grounded or (p.actionstate == 1)) and "Invisibility" or "Leaf Swirl"
			p.actionrings = (grounded or (p.actionstate == 1)) and 5 or 10
		end
		p.action2text = nil	-- this'll get changed later if we're invisible
		
		if (p.actionstate == 0)	-- starting the move?
			if (doaction == 1)	-- just pressed it
			and not (p.climbing)	-- we're not running up a wall
				B.PayRings(p)
				
				if (mo.espio_chargejump)	-- charge thrust
					local timer = mo.espio_chargejump_timer
					
					-- calculate speed, then launch!
					local range = (espio.ChargeThrustMax - espio.ChargeThrustMin)	-- we'll need the range for calcs
					local charge = min(espio.ChargeThrustMax, espio.ChargeThrustMin + ((range/espio.ChargeJumpDuration)*timer))
					
					local thrust = FixedMul(mo.scale, charge/((mo.eflags & MFE_UNDERWATER) and 2 or 1))
					mo.state = S_PLAY_ROLL
					mo.momz = 0	-- cancel z momentum for a full horizontal thrust
					P_InstaThrust(mo, mo.angle, thrust)
					S_StartSound(mo, sfx_kc5b)
					//print("Thrust: "..charge/FRACUNIT)
					
					-- apply cooldown
					local range = (espio.ChargeThrustMaxCooldown - espio.ChargeThrustMinCooldown)
					local cooldown = min(espio.ChargeThrustMaxCooldown, espio.ChargeThrustMinCooldown + ((timer*espio.ChargeJumpDuration)/range))
					B.ApplyCooldown(p, cooldown)
					
					-- visuals
					local dusts = 18
					for i = 1, dusts
						local radius, height = mo.radius>>FRACBITS, mo.height>>FRACBITS
						local x = P_RandomRange(-(radius), radius)*FRACUNIT
						local y = P_RandomRange(-(radius), radius)*FRACUNIT
						local z = P_RandomRange(0, height)*FRACUNIT
						
						local dust = P_SpawnMobjFromMobj(mo, x, y, z, MT_SPINDUST)
						if (mo.eflags & MFE_UNDERWATER) 
							dust.state = S_WPLD1 
							dust.scale = $/2
						end
					end
					mo.espio_chargejump_visualtics = espio.ChargeJumpVisualTics
					
					-- reset vars
					espio.StopChargeSounds(mo)
					espio.ResetChargeJump(mo)
				else
					p.actionstate = (grounded) and 1 or 2	-- invisibility is 1, leaf swirl is 2
					p.actiontime = 0	-- timer for abilities
					p.canguard = false
					
					-- init functions on each state
					if (p.actionstate == 1)	-- invisibility
						p.normalspeed = FixedMul(espio.InvisWalkMultiplier, skins[mo.skin].normalspeed)
						mo.espio_battleopacity = $ or 100
					else	-- leaf swirl
						p.pflags = $ & ~(PF_JUMPED|PF_SHIELDABILITY)
						p.secondjump = 0
						mo.state = S_ESPIO_LEAFSWIRL
						B.ApplyCooldown(p, espio.LeafSwirlCooldown)
					end
				end
			end
		else	-- ability thinker
			p.actiontime = $+1
			p.canguard = false
			
			if (p.actionstate == 1)	-- invisibility
				-- walk slower while invisible
				p.normalspeed = FixedMul(espio.InvisWalkMultiplier, skins[mo.skin].normalspeed)
				mo.espio_fixnormalspeed = true	-- we'll need this
				
				-- instantly go to waiting while invisible
				if (mo.state == S_PLAY_STND) then mo.state = S_PLAY_WAIT end
				
				-- drain rings while we're invisible
				if not (p.actiontime%espio.InvisRingDrainRate)
					P_GivePlayerRings(p, -1)
				end
				
				-- end conditions
				if not (doaction)	-- not holding attack anymore
					p.actionstate = 0
					B.ResetPlayerProperties(p)
					//B.ApplyCooldown(p, espio.InvisCooldown)
					
					mo.espio_removecooldown = 2
				end
			else	-- leaf swirl
				espio_leafSwirl(p)	-- easier to just have it in a function
			end
		end
		
		-- handle the leaf swirl hack
		if (mo.espio_leafswirl_parryhack)
			mo.espio_leafswirl_parryhack = $-1
		end
		
		-- get how opaque we should be based on momentum
		if (p.actionstate == 1)	-- we're in the invisibility state, adjust based on momentum!
			-- speed values
			local speedthreshold = FixedMul(p.normalspeed, espio.InvisWalkThreshold)/FRACUNIT
			local speed = FixedHypot(mo.momx, mo.momy)/FRACUNIT
			
			-- set how close we are to being visible as our action2text
			if (speed >= speedthreshold)
			or (p.rings == 0)	-- we're visible when we have no rings regardless
				p.action2text = "Visible!"
				p.action2textflags = (leveltime&4) and 2 or 0	-- TF_YELLOW = 2 (these arent exposed)
			else
				p.action2text = "Speed "..speed*100/speedthreshold.."%"
				p.action2textflags = 0
			end
			
			-- transparency ranges
			local mintrans = (p.rings == 0 and espio.InvisNoRingFade or 0) * 10
			local maxtrans = espio.InvisMaxFade * 10
			
			-- now get the target opacity
			local targetopacity = min(maxtrans, max(mintrans, 0))
			if (speed >= speedthreshold)
				speed = $ - speedthreshold
				local maxspeed = (p.normalspeed/FRACUNIT) - speedthreshold
				
				targetopacity = min(maxtrans, max(mintrans, (speed*maxtrans)/maxspeed))
			end
			
			if (mo.espio_battleopacity < targetopacity)
				mo.espio_battleopacity = min(targetopacity, $ + espio.InvisTransparencyAdjust)
			elseif (mo.espio_battleopacity > targetopacity)
				mo.espio_battleopacity = max(targetopacity, $ - espio.InvisTransparencyAdjust)
			end
		else
			-- might as well fix normalspeed here
			if (mo.espio_fixnormalspeed)
				B.ResetPlayerProperties(p)
				mo.espio_fixnormalspeed = nil
			end
			-- same for making sure invisibility has NO cooldown at all
			if (mo.espio_removecooldown)
				mo.espio_removecooldown = $-1
				if not (mo.espio_removecooldown)
					p.actioncooldown = 0
					mo.espio_removecooldown = nil
				end
			end
			
			if (mo.espio_battleopacity != nil)
				mo.espio_battleopacity = min(100, $ + espio.InvisTransparencyAdjust)
				if (mo.espio_battleopacity == 100)
					mo.espio_battleopacity = nil
					
					-- reset this while we're at it
					mo.frame = ($ & ~FF_TRANSMASK)
					mo.flags2 = $ & ~(MF2_DONTDRAW)
				end
			end
		end
		
		if (mo.espio_battleopacity != nil)
			-- now set our transparency!
			local trans = 10 - (mo.espio_battleopacity/10)
			//print("Transparency: "..trans)
			if (trans == 10)	-- fully transparent, use MF2_DONTDRAW
				mo.frame = ($ & ~FF_TRANSMASK) 	-- remove transparency, we wont need it here
				mo.flags2 = $|(MF2_DONTDRAW)
			else
				mo.frame = ($ & ~FF_TRANSMASK)
				mo.flags2 = $ & ~(MF2_DONTDRAW)
				if (trans >= 1 and trans <= 9)
					mo.frame = $|(trans << FF_TRANSSHIFT)
				end
			end
		end
	end
	
	-- espio exhaust
	local espio_battleExhaust
	if (B.VersionNumber:find("CE"))	-- community build has exhaust_func... BASED!!!!
		espio_battleExhaust = function(p)
			if not (espio.AbilityCheck(p)) then return end
			local mo = p.mo
			
			-- conditions to lower it
			if (mo.espio_ceilingCling)
			or (mo.espio_wallrun)
				local maxtime = espio.BattleExhaustTime
				mo.espio_exhaustmeter = max(0,$-FRACUNIT/maxtime)	-- we'll use our own var so we dont conflict
				p.exhaustmeter = mo.espio_exhaustmeter
				
				-- what to do when we've exhausted ourself
				if (p.exhaustmeter <= 0)
				and (p.climbing)
					mo.espio_wallrun_timer = 0	-- no more running up
				end
			end
			-- refill here
			if ((P_IsObjectOnGround(mo) and not mo.espio_ceilingCling) or P_PlayerInPain(p))
				mo.espio_exhaustmeter = FRACUNIT
				p.exhaustmeter = FRACUNIT
			end
	
			return true
		end
	else -- oh boy
		local homingthok = function(p)
		-- 	if not (p and p.valid and p.charability == CA_HOMINGTHOK) then return end
		-- 	if not (p.pflags&PF_SHIELDABILITY)
				p.homing = min($,10)
		-- 	end
		end

		local popgun = function(player)
			if P_PlayerInPain(player) and player.charability2 == CA2_GUNSLINGER then
				player.weapondelay = TICRATE/3
			end
		end

		local weight = function(player)
			if not(player.mo) then return end
				
			local w
			if S[player.skinvars].weight then
				w = S[player.skinvars].weight
			else
				w = S[-1].weight
			end
			
			player.mo.weight = FRACUNIT*w/100
		end

		local exhaust = function(player)
			if player.eggrobo_transforming then return end
			local mo = player.mo
			if mo.exhaustcolor == nil then
				mo.exhaustcolor = false
			end
			if ((P_IsObjectOnGround(mo) and not mo.espio_ceilingCling) or P_PlayerInPain(player))	-- ESPIO CODE
			and (not player.actionstate and not espio.AbilityCheck(player))	-- ESPIO CODE
			and player.exhaustmeter != FRACUNIT 
			and not(S[player.skinvars].special == B.Action.Dig and player.mo.flags&MF_NOCLIPTHING)
				//Refill meter
				player.exhaustmeter = FRACUNIT
				if mo.exhaustcolor == true then
					mo.colorized = false
					mo.color = player.skincolor
					mo.exhaustcolor = false
				end
			return end
			local use = false
			local maxtime = TICRATE*6
			local warningtic = FRACUNIT/3
			if (player.powers[pw_carry] == CR_MACESPIN and player.mo.tracer) //Swinging from chain
				maxtime = TICRATE*6
				use = true
			elseif player.charability == CA_FLY then //Tails ability
				maxtime = TICRATE*11/2
			-- ESPIO CODE
			elseif (espio.AbilityCheck(player))
				maxtime = espio.BattleExhaustTime
			-- ESPIO CODE END
			elseif player.charability == CA_GLIDEANDCLIMB and not(player.mo.flags&MF_NOCLIPTHING) then //Knuckles ability
				maxtime = $+TICRATE*2
			end
			//Get usage
			-- ESPIO CODE
			if (espio.AbilityCheck(player))
				if (mo.espio_ceilingCling)
				or (mo.espio_wallrun)
					use = true
				end
			-- ESPIO CODE END
			elseif
				player.powers[pw_tailsfly]
				or
				player.climbing
				or (player.charability==CA_FLOAT and (
					player.secondjump or player.pflags&(PF_THOKKED|PF_JUMPED)==PF_THOKKED|PF_JUMPED)
				)
				use = true
			end
			//Use meter
			if(use) and (player.exhaustmeter) then
				if player.climbing == 5 then
					if player.exhaustmeter > warningtic then
						player.exhaustmeter = max(warningtic,$-FRACUNIT/4)
					else
						player.exhaustmeter = max(0,$-FRACUNIT/4)
					end
				end
				player.exhaustmeter = max(0,$-FRACUNIT/maxtime)
			end
			//Exhaust warning
			local colorflip = false
			if player.exhaustmeter < warningtic and not(player.exhaustmeter == 0) then
				if not(leveltime&7)
		-- 			S_StartSound(mo,sfx_s3kb5,player)
					S_StartSound(mo,sfx_s3kbb,player)
				end
				if (leveltime&4)
					colorflip = true
				end
			end
				
			//Exhaust color
			if colorflip == true
				mo.exhaustcolor = true
				mo.colorized = true
				mo.color = SKINCOLOR_BRONZE
			elseif mo.exhaustcolor == true
				mo.exhaustcolor = false
				mo.colorized = false
				mo.color = player.skincolor		
			end
				
			//Fully exhausted
			if use and not(player.exhaustmeter) then
				//Tails
				if player.powers[pw_tailsfly] then
					player.powers[pw_tailsfly] = 0
				end
				//Knuckles
				if player.climbing then
					-- ESPIO CODE
					if (espio.AbilityCheck(player))
						mo.espio_wallrun_timer = 0
					else
					-- ESPIO CODE END
						player.climbing = 0
						player.pflags = $|PF_JUMPED|PF_THOKKED
						mo.state = S_PLAY_ROLL
					end
				end
				//Metal Sonic
				if (player.charability==CA_FLOAT and player.secondjump) then
					local spd = FixedHypot(player.rmomx,player.rmomy)
					local ang = R_PointToAngle2(0,0,player.rmomx,player.rmomy)
					local limit = FixedMul(player.normalspeed/6,mo.scale)
					if spd > limit then
						P_InstaThrust(mo,ang,limit)
						if P_RandomChance(FRACUNIT/6) then
							local x = mo.x-P_ReturnThrustX(mo,mo.angle,mo.radius)
							local y = mo.y-P_ReturnThrustY(mo,mo.angle,mo.radius)
							local z = mo.z+mo.height/2
							P_SpawnMobj(x,y,z,MT_SMOKE)
						end
						if not(S_SoundPlaying(mo,sfx_s3kc2s)) then
							S_StartSound(mo,sfx_s3kc2s)
						end
					end
				end
				//Swinging from chain
				if (player.powers[pw_carry] == CR_MACESPIN and player.mo.tracer)
					player.powers[pw_carry] = CR_NONE
					player.mo.tracer = nil
					player.powers[pw_flashing] = TICRATE/4
				end
			end
		end

		B.CharAbilityControl = function(player)
			homingthok(player)
			popgun(player)
			weight(player)
			exhaust(player)
			B.ShieldAbility(player)
		end
	end
	
	-- setup espio's skinvars
	S["espio"] = {
		weight = 90,
		special = espio_battleSpecial,
		func_priority_ext = espio_battlePriority,
		func_exhaust = espio_battleExhaust,	-- CE based
		guard_frame = 6,
	}
	
	loaded = true
end

-- hide overlays while we're invisible
-- (gotta do this outside the function since you cant add hooks within loops/coroutines)
local function espio_hideOverlay(overlay)
	local owner = overlay.target
	if not (owner and owner.valid) then return end
	if not (owner.player) then return end	-- this is needed for MT_OVERLAY apparently
		
	if (owner.espio_battleopacity != nil) 
		overlay.flags2 = $|(MF2_DONTDRAW)
		if (overlay.tracer and overlay.tracer.valid)
			-- the overlay has a 2nd layer (like the s3k shields!)
			overlay.tracer.flags2 = $|(MF2_DONTDRAW)
		end
		
		return true 
	end
	
	-- if we're here, we have no need for it to not be drawn
	overlay.flags2 = $ & ~(MF2_DONTDRAW)
	if (overlay.tracer and overlay.tracer.valid)
		overlay.tracer.flags2 = $ & ~(MF2_DONTDRAW)
	end
	
	return false
end
local overlays = {
MT_ELEMENTAL_ORB, MT_FORCE_ORB, MT_ATTRACT_ORB, MT_ARMAGEDDON_ORB, MT_WHIRLWIND_ORB, MT_PITY_ORB, 
MT_FLAMEAURA_ORB, MT_BUBBLEWRAP_ORB, MT_THUNDERCOIN_ORB, 
//MT_OVERLAY	-- needs its own thinker for this
}
for i = 1, #overlays
	addHook("MobjThinker", espio_hideOverlay, overlays[i])
end
addHook("MobjThinker", function(overlay)
	if (overlay.target and overlay.target.valid)
	and (overlay.target.flags2 & MF2_SHIELD)	-- this is a 2nd layer for a shield
		-- we'll handle this inside that thinker instead
		return
	end
	
	-- now just run this like normal
	espio_hideOverlay(overlay)
end, MT_OVERLAY)