-- Everything related to colour shifting
espio.ColourShift = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if (mo.skin != "espio") then mo.espio_shifttimer = nil return end
	
	-- conditions where we should *not* colour shift
	if ((gametyperules & GTR_TEAMS) and p.ctfteam)	-- no confusion in team gamemodes
	or (mo.frame & FF_TRANSMASK) -- multiple ghosts stacking will make us more opaque, so lets prevent it by removing it outright
		mo.espio_shifttimer = nil
		return
	end
	
	-- conditions when we should start colour shifting
	if ((mo.state == S_PLAY_WAIT) or (mo.state == S_PLAY_NIGHTS_STAND)	-- we're bored or standing in nights
	and (mo.frame > 0))	-- and we're not in the transition frame
	//or (mo.espio_focuscharges == 5)	-- we're fully charged!
	or (mo.espio_chargejump_timer and mo.espio_chargejump_timer > espio.ChargeJumpDuration)	-- we're fully charged!
	or (mo.state == S_PLAY_EDGE)	-- we're on the edge
	or (Tether and mo.state == S_PLAY_TETHER_HOLD)	-- smart tether support!
		if (mo.espio_colourshift_delay)
			mo.espio_colourshift_delay = $-1
		else
			mo.espio_shifttimer = $ or 0
		end
	end
	
	if (mo.espio_shifttimer != nil)	-- if it's not nil, we can start
		local startColours = {mo.color, SKINCOLOR_SAPPHIRE, SKINCOLOR_APPLE, SKINCOLOR_YELLOW, SKINCOLOR_SALMON}	-- the colours we start from
		local shiftColours = {SKINCOLOR_SAPPHIRE, SKINCOLOR_APPLE, SKINCOLOR_YELLOW, SKINCOLOR_SALMON, mo.color}	-- the colours we shift to
		
		-- handle the timer
		mo.espio_shifttimer = $+1
		
		-- the start colour, what we're fading FROM
		local start = P_SpawnGhostMobj(mo)
		start.frame = ($ & ~FF_TRANSMASK)
		start.rollangle = mo.rollangle
		start.color = startColours[((mo.espio_shifttimer/20)%5)+1]
		start.tics = 1
		
		-- the shift colour, what we're fading TO
		local fade = 10 - ((mo.espio_shifttimer%20)/2)
		
		local shift	-- define here so it's usable outside the block
		if (fade >= 1 and fade <= 9)
			shift = P_SpawnGhostMobj(mo)
			shift.frame = ($ & ~FF_TRANSMASK) | (fade << FF_TRANSSHIFT)
			shift.rollangle = mo.rollangle
			shift.color = shiftColours[((mo.espio_shifttimer/20)%5)+1]
			shift.tics = 1
		end
		
		if (mo.flags2 & MF2_DONTDRAW)	-- hide them if we arent visible
			-- we want to hide them instead of preventing them from being spawned to prevent desync
			start.flags2 = $|MF2_DONTDRAW
			if (shift and shift.valid)
				shift.flags2 = $|MF2_DONTDRAW
			end
		end
		
		if (mo.espio_shifttimer == ((20*#startColours)-1))
			mo.espio_shifttimer = nil
		end
	end
end