-- Everything related to Espio's Wallrun
espio.resetWallRun = function(p)
	local mo = p.mo
	
	p.espio_wallrun, mo.espio_wallrun = nil, nil
	mo.espio_wallrun_lastmomz = nil
	mo.espio_wallrun_slidemomz = nil
	mo.espio_wallrun_leveltime = nil
end

local function espio_clingDelay(p)
	local mo = p.mo
	
	-- handle the delay before we can cling onto a wall
	if (P_IsObjectOnGround(mo))
		-- we only wanna reset this on the ground
		mo.espio_wallrun_clingdelay = 16
	else
		mo.espio_wallrun_clingdelay = $ and $-1
	end
end

local function espio_checkLine(p, line, side)
	local mo = p.mo
	
	-- might as well do this here
	if (CBW_Battle and p.exhaustmeter == 0) then return false end	-- we're exhausted (battle mod)
	if (p.gotflag or p.gotcrystal) then return false end	-- we have the flag (or the crystal in battlemod)
	
	-- really easy one, dont cling to noclimb walls
	if (line.flags & ML_NOCLIMB) then return false end	
	
	-- sector checks
	if not (line.backsector) then return true end	-- apparently in these situations source always returns true?
	local sec = (line.backsector and not P_PointOnLineSide(mo.x, mo.y, line)) and line.backsector or line.frontsector
	//print((sec == line.backsector) and "backsector" or "frontsector")
	
	local canclimb = false
	local checkz = {mo.z, mo.z + mo.height}
	if (mo.eflags & MFE_VERTICALFLIP) then checkz = {mo.z + mo.height, mo.z} end
	
	-- floor check
	//print("Floor CheckZ: "..checkz[2])
	//print("Floor Height: "..sec.floorheight/FRACUNIT)
	if (checkz[2] < sec.floorheight) 
		canclimb = true 
		//print("floor check") 
	end
	
	-- ceiling checks
	//print("Ceiling CheckZ: "..checkz[1])
	//print("Ceiling Height: "..sec.ceilingheight/FRACUNIT)
	local thokbarrier = (sec.ceilingpic:upper():find("F_SKY"))	-- walls above this ceiling arent rendered, shouldnt be able to climb?
	if ((checkz[1] >= sec.ceilingheight) and not thokbarrier) 
		canclimb = true
		//print("ceiling check") 
	end
	
	-- FOF checks
	for rover in sec.ffloors()
		if not (rover.flags & FF_EXISTS or rover.flags & FF_BLOCKPLAYER) then continue end
		
		if (checkz[1] > rover.bottomheight)
		and (checkz[2] < rover.topheight)
			canclimb = true
			//print("FOF check")
			break
		end
	end
	
	-- if we're here, line checks went right
	return canclimb
end

-- this wall detection is my greatest achievement
hookLib.AddHook("PlayerLineCollide", function(p, l, s)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then return end
	if (mo.espio_wallrun) then return end	-- we're already running up a wall!
	
	if (p.pflags & PF_JUMPED)	-- we need to be jumping
	and (p.cmd.buttons & BT_JUMP)	-- we also need to be holding jump
	and not (mo.espio_wallrun_clingdelay)	-- need to be jumping for long enough
	and (espio_checkLine(p, l, s))	-- make sure everything is good here
		-- face the wall
		local whichside = P_PointOnLineSide(mo.x, mo.y, l)
		local angle = R_PointToAngle2(l.v1.x, l.v1.y, l.v2.x, l.v2.y)
		angle = $ + (ANGLE_90 * (whichside and -1 or 1))
				
		mo.angle = angle
		p.drawangle = angle
			
		-- make us start running up it (this is a dirty climbing hack)
		p.charability = CA_GLIDEANDCLIMB
		p.climbing = 5
		p.espio_wallrun, mo.espio_wallrun = true, true	-- trust me there's a reason we have two saved
		mo.espio_wallrun_lastmomz = mo.momz
		mo.espio_wallrun_leveltime = leveltime
		S_StartSound(mo, sfx_kc45)
			
		-- shield stuff
		p.pflags = $ & ~(PF_SHIELDABILITY)
		p.secondjump = 0
			
		-- increase timer
		mo.espio_wallrun_timer = min(FRACUNIT, $ + espio.WallRunAdd)
	end
end)

espio.WallRun = function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	if not (espio.AbilityCheck(p)) then espio.resetWallRun(p) return end
	
	if (P_IsObjectOnGround(mo)) then mo.espio_wallrun_timer = FRACUNIT end
	-- ^ reset timer when grounded
	
	espio_clingDelay(p)
	//espio_clingWalls(p) -- now handled with hookLib for compatibility with other scripts
	-- ^ cling to walls
	
	if (p.climbing)
		-- lower stamina
		-- forward check
		local forward
		if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
			forward = (p.cmd.sidemove > 0 and mo.angle == 0 or p.cmd.sidemove < 0 and mo.angle == ANGLE_180)
		else
			forward = p.cmd.forwardmove > 0
		end
		
		local ticsremove = (FRACUNIT/espio.WallRunTics)/(forward and 1 or 3)
		mo.espio_wallrun_timer = max(0, $ - ticsremove)
		
		if not (p.cmd.buttons & BT_JUMP)	-- if we're not holding jump, get off the wall!
			-- jump off
			p.charability = CA_ESPIOREAL
			p.climbing = 0
			P_DoJump(p, true)
			p.pflags = $ & ~(PF_STARTJUMP|PF_THOKKED)
			mo.angle = $ + ANGLE_180
			P_InstaThrust(mo, mo.angle, FixedMul(p.jumpfactor, FixedMul(mo.scale, espio.WallRunJumpSpeed)))
			S_StartSound(mo, sfx_jump)
			
			-- momz is a lot more annoying than I thought
			mo.momz = FixedMul(p.jumpfactor, mo.espio_wallrun_lastmomz)
			
			-- minimum height
			local minheight = FixedMul(p.jumpfactor, FixedMul(mo.scale, espio.WallRunJumpMinHeight))
			
			if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > -(minheight))
			or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < minheight)
				mo.momz = (minheight)*P_MobjFlip(mo)
			end
			espio.resetWallRun(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		end
	else	-- make sure our ability is reset if we're not wall running anymore
		if (p.espio_wallrun)
			if (mo.espio_wallrun)	-- THIS is why we had both saved
				if not (P_IsObjectOnGround(mo))
					mo.momz = mo.espio_wallrun_lastmomz*3/4
					
					if (p.cmd.buttons & BT_SPIN)
						-- minimum height
						if ((mo.eflags & MFE_VERTICALFLIP) and mo.momz > FixedMul(mo.scale, -3*FRACUNIT))
						or (not (mo.eflags & MFE_VERTICALFLIP) and mo.momz < FixedMul(mo.scale, 3*FRACUNIT))
							mo.momz = FixedMul(mo.scale, FRACUNIT*3) * P_MobjFlip(mo)
						end
					else
						p.pflags = $|(PF_STARTJUMP) & ~(PF_THOKKED)
						mo.espio_wallrun_clingdelay = 16
					end
				end
				
				S_StartSound(mo, sfx_kc45)
			end
			
			-- this needs to always run, otherwise we can keep the wrong ability through levels
			p.charability = CA_ESPIOREAL
			espio.resetWallRun(p)
			
			-- reset these to prevent reclings
			p.lastlinehit = -1
			p.lastsidehit = -1
			return
		end
	end
end

addHook("PreThinkFrame", do
	for p in players.iterate
		local mo = p.mo
		
		if not (espio.AbilityCheck(p)) then continue end
		if not (p.climbing) then continue end
		
		-- slight hack to make horizontal momentum capped
		-- setting forward move to be -1/1 max so we can still have wall scrolling momentum without too much power
		local forward
		if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
			forward = (p.cmd.sidemove > 0 and mo.angle == 0 or p.cmd.sidemove < 0 and mo.angle == ANGLE_180)
		else
			forward = p.cmd.forwardmove > 0
		end
		if (forward) then p.cmd.forwardmove = min(1, $) continue end
		
		p.cmd.forwardmove = -1
		p.cmd.sidemove = max(-25, min(25, $))	-- limit side wall sliding momentum
	end
end)

-- playerthink allows us to edit momentum while climbing
addHook("PlayerThink", function(p)
	local mo = p.mo
	if not (mo and mo.valid) then return end
	
	-- wall running checks
	if not (espio.AbilityCheck(p)) then return end
	if not (p.climbing) then return end
	
	local wallruntimer = leveltime - mo.espio_wallrun_leveltime
	-- ^ add some difference between multiple espios
		
	-- forward check
	local forward
	if (twodlevel or (mo.flags2 & MF2_TWOD))	-- fuck
		forward = (p.cmd.sidemove > 0 and mo.angle == 0 or p.cmd.sidemove < 0 and mo.angle == ANGLE_180)
	else
		forward = p.cmd.forwardmove > 0
	end
	
	-- give a speed boost when you're super or have super sneakers
	local speedup = (p.powers[pw_sneakers] or P_PlayerFullbright(p))
	
	-- detect if we're moving forward
	if (forward)
	and (mo.espio_wallrun_timer)	-- we need stamina too
		local tics = (mo.espio_wallrun_timer*100)/FRACUNIT
		local maxspeed = FixedMul(p.normalspeed, espio.WallRunSpeedModifier)
		local speed = (tics*(maxspeed/FRACUNIT)/100)*FRACUNIT / ((mo.eflags & MFE_UNDERWATER) and 2 or 1)
		if (speedup) then speed = $*3/2 end
		
		mo.momz = $ + FixedMul(mo.scale, (speed*P_MobjFlip(mo)))
		mo.espio_wallrun_lastmomz = mo.momz
		mo.espio_wallrun_slidemomz = nil
		
		-- effects
		if (mo.state != S_PLAY_CLIMB) then mo.state = S_PLAY_CLIMB end
		espio.SetFrame(mo, (speedup or (mo.espio_wallrun_timer < espio.WallRunTiredThreshold)) and (wallruntimer%4) or (wallruntimer%8)/2, SPR2_CLMB)
		-- getting tired? extra effects
		if (mo.espio_wallrun_timer < espio.WallRunTiredThreshold)
			if ((wallruntimer%4) == 1 or (wallruntimer%4) == 3)
				for i = 1, 2
					-- distance to the wall
					local distance = 13
					local xoff = distance*cos(mo.angle)
					local yoff = distance*sin(mo.angle)
					
					-- make the dust come from espio's shoes
					local angle = mo.angle - (((wallruntimer%4) == 1) and ANGLE_270 or ANGLE_90)
					local distance = 14 + P_RandomRange(-8, 8)
					xoff = $ + distance*cos(angle)
					yoff = $ + distance*sin(angle)
					
					local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, P_RandomRange(-2, 2)*FRACUNIT, MT_SPINDUST)
					dust.scale = $*2/3
					if (mo.eflags & MFE_UNDERWATER)
						dust.state = S_SPINDUST_BUBBLE1
						S_StartSoundAtVolume(mo, _G["sfx_bubbl"..P_RandomRange(1, 5)], 128)
					else
						S_StartSoundAtVolume(mo, sfx_s251, 96)
					end
					
					-- make the dust burst out a bit from us
					local momz = 2*FRACUNIT + (P_RandomFixed()<<2)
					local angle = mo.angle + (ANG1*P_RandomRange(-35, 35))
					local speed = (FRACUNIT/4) + (P_RandomFixed()<<3)
					
					P_SetObjectMomZ(dust, momz)
					P_InstaThrust(dust, angle, -(speed))
					P_TryMove(dust, dust.x + dust.momx, dust.y + dust.momy, true)
				end
			end
		end
	else	-- slide down
		if (mo.espio_wallrun_slidemomz == nil)
			-- assume this is when we start sliding
			mo.espio_wallrun_slidemomz = mo.espio_wallrun_lastmomz/2
		end
		
		-- modify slide momz
		local removemomz = FixedMul(mo.scale, espio.WallRunSpeedSlide)*P_MobjFlip(mo)
		-- flip this again in goop
		if (mo.eflags & MFE_GOOWATER)
			removemomz = $ * -1
		end
		
		mo.espio_wallrun_slidemomz = $ - (removemomz / ((mo.eflags & MFE_UNDERWATER) and 2 or 1))
		
		if (mo.state != S_PLAY_CLING) then mo.state = S_PLAY_CLING end
		espio.SetFrame(mo, 0, SPR2_CLNG)
		mo.momz = $ + mo.espio_wallrun_slidemomz
		mo.espio_wallrun_lastmomz = mo.momz
		
		-- spawn dust!
		if not (wallruntimer%3)
			for i = 1, 2
				-- distance to the wall
				local distance = 13
				local xoff = distance*cos(mo.angle)
				local yoff = distance*sin(mo.angle)
				
				-- distance to espio's hand (with some randomization)
				local distance = 15 + P_RandomRange(-8, 8)
				xoff = $ + distance*cos(mo.angle - ANGLE_90)
				yoff = $ + distance*sin(mo.angle - ANGLE_90)
				local zoff = FRACUNIT*42 + (P_RandomRange(-8, 8)*FRACUNIT)
				
				local dust = P_SpawnMobjFromMobj(mo, xoff, yoff, zoff, MT_SPINDUST)
				dust.state = (mo.eflags & MFE_UNDERWATER) and S_SPINDUST_BUBBLE1 or $
				dust.scale = $*2/3
				S_StartSound(mo, sfx_s3k7e)
			end
		end
	end
	
	-- visual ghost (copypaste from source)
	if (speedup) and (abs(mo.momz) > FixedMul(mo.scale, FRACUNIT*10))
		local g = P_SpawnGhostMobj(mo)
		g.fuse = 2
		if (g.tracer) then g.tracer.fuse = 2 end
			
		if (leveltime%2)
			g.frame = $ & ~(FF_TRANSMASK)
			g.frame = $|FF_TRANS70
			if (g.tracer)
				g.tracer.frame = $ & ~(FF_TRANSMASK)
				g.tracer.frame = $|(FF_TRANS70)
			end
		end
	end
end)